/***********************************************************************
 * userfun.cpp: userdefined functions
 ***********************************************************************/
/*
  Defining new functions:

   - add the function name and argument count to RegisterFunctions(), e.g
       if (ifun == 0) {
         strcpy(cnam,"myselect");
         *iargc = 3;
       }
   
   - implement the function at the end of this source file using pop() and push()
     for input and output signal interface, e.g. 
       void __stdcall myselect(void){ ... }
   
   - add the function name and ordinal number to the export file digexfun.def, e.g.
       myselect @2

*/

#include <stdio.h>
#include <math.h> 
#include <stdlib.h> 
#include <string.h> 
#include <ctype.h> 

#include "digexfun.hpp"

int __stdcall RegisterFunctions(int ifun, char* cnam, int* iargc)
{ // register userdefined functions number ifun:
  //   return name (cnam[50]) and number of arguments (iargc)
  // return != 0, if ifun exceeds the number of available functions
  if (!cnam || !iargc) return 1;

  LoadDigsiLibrary();

  if (ifun == 0) {
    strcpy(cnam,"myselect");
    *iargc = 3;
  
  } else if (ifun == 1) {
    strcpy(cnam,"LIM");
    *iargc = 3;
  
  } else if (ifun == 2) {
    strcpy(cnam,"SwitchScheme");
    *iargc = -1; // variable No of arguments
 
  } else if (ifun == 3) {
    strcpy(cnam,"GetSwitchAction");
    *iargc = 2;
  
  } else { // unknown function
    return 1;
  }

  return 0;
}

void __stdcall myselect(void)
{ // simple select without interrupt
  // arguments: dbool, x, y
  
  // pop input signals from stack
  double dbool = pop(); // 1. argument
  double x = pop();     // 2. argument
  double y = pop();     // 3. argument
  
  // push output signals to stack
  if (dbool > 0.5) {
    push(x);
  } else {
    push(y);
  }
}

void __stdcall LIM(void)
{ // simple lim without interrupt
  // arguments: input u, min, max
  
  // pop input signals from stack
  double u   = pop(); // 1. argument
  double min = pop(); // 2. argument
  double max = pop(); // 3. argument
  
  // push output signal to stack
  if (u < min) {
    push(min);
  } else if (u > max) {
    push(max);
  } else {
    push(u);
  }
}
/*
void __stdcall GetBit(void)
{ // return one bit of input u as double
  // arguments: input u (bitfield), bit No starting with 0
  
  // pop input signals from stack
  double du   = pop(); // 1. argument
  double dbit = pop(); // 2. argument
  
  int iu = *(int*)((void*)&du); // interpretation of double as int
  int ibit = int(dbit+0.5);

  int ires = (iu >> ibit) & 1;
  // push output signal to stack
  push((double) ires);
}

void __stdcall SetBit(void)
{ // return a certain bit as bitfield
  // arguments: input bit No starting with 0
  
  // pop input signals from stack
  double dbit = pop(); // 1. argument
  
  int ibit = int(dbit+0.5);

  int ires = (1 << ibit);
  // push output signal to stack
  push( *((double*) ((void*)&ires)) );
}
*/
#define MYROUND(a) ( (a) > 0 ? (floor((a)+0.5)) : (ceil((a)-0.5)) )
#define MAXNO 64

class SwitchProblem
{
public:
  double dSize[MAXNO];   // size of switch element
  int iContr[MAXNO];     // element controlled (1=on, 0=off)
  int iStatus[MAXNO];    // present switch status per element (-1=off, 1=on)
  int iStatusNew[MAXNO]; // temp. switch status per element (-1=off, 1=on)
  int iResStatus[MAXNO]; // proposed switch status per element (-1=off, 1=on)
  int iMode;             // required switching action: -1 off, 0 no action, 1 on
  int iMax;              // present max. of elements

  SwitchProblem() { Reset();
  }
  virtual ~SwitchProblem() {  }

  void Reset(){ 
    memset(dSize,0,sizeof(double)*MAXNO);
    
    memset(iContr,0,sizeof(int)*MAXNO);
    memset(iStatus,0,sizeof(int)*MAXNO);
    memset(iStatusNew,0,sizeof(int)*MAXNO);
    memset(iResStatus,0,sizeof(int)*MAXNO);
    
    iMode = 0; iMax = 0;
  }
  int GetResStatus(int iElm){ return iResStatus[iElm]; }

  double GetSumOfSize(int *iStat)
  { // return sum of all controlled elements which are switched on (status iStat)
    // input iStatus: array of switch status
    double dSum = 0; // result
    if (iStat) {
      for(int i=0; i<iMax; i++) { // for each controlled element
        if (iContr[i]) { // controlled
          if (iStat[i] == 1) { // on
            dSum += dSize[i];
          }
        }
      }
    }
    return dSum;
  }


  int SwitchOn()
  { // switch next element(s) on (iStatusNew), stop when size increases 
    // return 0 if next statuses found, 1 if limit exceeded (all elements switched on)

    // find smallest element which is switched off
    int ifound = -1;
    for(int i=0; i<iMax && ifound==-1; i++) { // for each controlled element
      if (iContr[i]) { // controlled
        if (iStatusNew[i] == -1) { // off
          ifound = i;
        }
      }
    }
    if (ifound == -1) { // all elements on
      return 1;
    }
    
    double dSumInit = GetSumOfSize(iStatus); // original size

    // switch this element on
    iStatusNew[ifound] = 1;
 
    // switch all smaller elements off
    for(i=0; i<ifound; i++) { // for each smaller and controlled element
      if (iContr[i]) { // controlled
        iStatusNew[i] = -1; // off
      }
    }

    double dSum = GetSumOfSize(iStatusNew);
    if (dSum <= dSumInit) { // not yet bigger
      return SwitchOn(); // recursive call
    }

    return 0;
  }

  int SwitchOff()
  { // switch next element(s) off (iStatusNew), stop when size decreases 
    // return 0 if next statuses found, 1 if limit exceeded (all elements switched off)

    // find smallest element which is switched on
    int ifound = -1;
    for(int i=0; i<iMax && ifound==-1; i++) { // for each controlled element
      if (iContr[i]) { // controlled
        if (iStatusNew[i] == 1) { // on
          ifound = i;
        }
      }
    }
    if (ifound == -1) { // all elements off
      return 1;
    }
    
    double dSumInit = GetSumOfSize(iStatus); // original size

    // switch this element off
    iStatusNew[ifound] = -1;
 
    // switch all smaller elements on
    for(i=0; i<ifound; i++) { // for each smaller and controlled element
      if (iContr[i]) { // controlled
        iStatusNew[i] = 1; // on
      }
    }

    double dSum = GetSumOfSize(iStatusNew);
    if (dSum >= dSumInit) { // not yet smaller
      return SwitchOff(); // recursive call
    }

    return 0;
  }

  void DetermineResultingActions()
  { // compare iStatusNew and iStatus and set iResStatus
    for(int i=0; i<iMax; i++) { // for each controlled element
      if (iContr[i]) { // controlled
        if (iStatusNew[i] != iStatus[i]) { // action required, status changed
          iResStatus[i] = iStatusNew[i];
        }
      }
    }
  }

};

static SwitchProblem aSP;


void __stdcall SwitchScheme(void)
{ // determine switching action which is stored in aSP
  // arguments: input n, No of controlled elements
  //            input mode, required switching action: -1 off, 0 no action, 1 on
  //            for each controlled element 3 inputs: size, controlled, present switch status
  // The controlled elements have to be sorted by size beginning with the smallest.
 
  int iRes = 0;

  aSP.Reset(); // initialize static

  // pop input signals from stack
  double dNo = pop(); // 1. argument, No of controlled elements
  aSP.iMax = int(MYROUND(dNo));

  double dmode = pop(); // 2. argument, required switching action: -1 off, 0 no action, 1 on
  aSP.iMode = int(MYROUND(dmode));

  for(int i=0; i<aSP.iMax; i++) { // for each controlled element
    aSP.dSize[i] = pop(); // next argument: size, e.g. shunt size in Mvar


    double dContr = pop(); // next argument, controlled, yes = 1, 0 = no
    aSP.iContr[i] = int(MYROUND(dContr));

    double dStatus = pop(); // next argument, present switch status,  -1 off, 1 on
    aSP.iStatus[i] = int(MYROUND(dStatus));
    aSP.iStatusNew[i] = aSP.iStatus[i];
  }

  double dSum = aSP.GetSumOfSize(aSP.iStatus); // present sum
  
  if (aSP.iMode == 0) { // no action required
    push(dSum); // push present sum of size
    return;
  }

  if (aSP.iMode == 1) { // action on
    aSP.SwitchOn();
  } else if (aSP.iMode == -1) { // action off
    aSP.SwitchOff();
  }

  aSP.DetermineResultingActions();
  
  // dSum = aSP.GetSumOfSize(aSP.iStatusNew);
  push(dSum); // push present sum of size
}

void __stdcall GetSwitchAction(void)
{ // return action for certain switch, SwitchScheme() has to be called first 
  // arguments: input u (sum of size), not used but required for DSL parser to get correct equation sequence
  //            input switch No
  
  // pop input signals from stack
  double du  = pop(); // 1. argument
  double dNo = pop(); // 2. argument
  
  int iNo = int(dNo+0.5);

  // push output signal to stack
  push((double) aSP.GetResStatus(iNo));
}
